/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  T z C L O C K . C                                                                                 *
 *  =================                                                                                 *
 *                                                                                                    *
 *  TzClock developed by Chris Knight based on glock by Eric L. Smith.                                * 
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  TzClock is free software; you can redistribute it and/or modify it under the terms of the GNU     *
 *  General Public License version 2 as published by the Free Software Foundation.  Note that I       *
 *  am not granting permission to redistribute or modify TzClock under the terms of any later         *
 *  version of the General Public License.                                                            *
 *                                                                                                    *
 *  TzClock is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without      *
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 *  GNU General Public License for more details.                                                      *
 *                                                                                                    *
 *  You should have received a copy of the GNU General Public License along with this program (in     * 
 *  the file "COPYING"); if not, write to the Free Software Foundation, Inc.,                         *
 *  59 Temple Place - Suite 330, Boston, MA 02111, USA.                                               *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @file
 *  @brief .
 */
 
#include "config.h"
#include "TzClockDisp.h"

extern const guint gtk_major_version;
extern const guint gtk_minor_version;
extern const guint gtk_micro_version;

#if (GTK_MAJOR_VERSION == 2 && GTK_MINOR_VERSION >= 10) || GTK_MAJOR_VERSION > 2
	#define STATUS_ICON
#endif

char *nameColours[MAX__COLOURS] =
{
	"blk", "wht", 
	"cbf", "cbn", "cbi", "fce", "fsw", "txt", "hrh", 
	"mnh", "sch", "alh", "ctc", "hrm", "mnm", "qum", 
	"qmf", "hhf", "mnf", "ccf", "swh", "swc", "swf", 
	"swm"
};

/*----------------------------------------------------------------------------------------------------*
 * Default colour tables for the basic light and dark colour options, overridden by command line.     *
 *----------------------------------------------------------------------------------------------------*/
char delaultColour[MAX__COLOURS][10] =		/* Default colours */
{
#ifdef COLOUR_GREEN
	/* 00 BLK */ "#000000", /* 01 WHT */ "#FFFFFF",
	/* 02 CBF */ "#60FF60", /* 03 CBN */ "#C0C0C0", /* 04 CBI */ "#608060", /* 05 FCE */ "#004000",
	/* 06 FSW */ "#000000", /* 07 TXT */ "#60FF60", /* 08 HRH */ "#E0E0E0", /* 09 MNH */ "#E0E0E0", 
	/* 10 SCH */ "#FF0000", /* 11 ALH */ "#000080",	/* 12 CTC */ "#FF0000", /* 13 HRM */ "#A0A0A0", 
	/* 14 MNM */ "#000000", /* 15 QUM */ "#C0C000", /* 16 GMF */ "#A0A000", /* 17 HHF */ "#606060", 
	/* 18 MNF */ "#A0A0A0", /* 19 CCF */ "#E00000", /* 20 SWH */ "#FF0000", /* 21 SWC */ "#FF0000", 
	/* 22 SWF */ "#E00000", /* 23 SWM */ "#60FF60"
#else
	/* 00 BLK */ "#000000", /* 01 WHT */ "#FFFFFF",
	/* 02 CBF */ "#6060FF", /* 03 CBN */ "#C0C0C0", /* 04 CBI */ "#606080", /* 05 FCE */ "#000040",
	/* 06 FSW */ "#000000", /* 07 TXT */ "#6060FF", /* 08 HRH */ "#E0E0E0", /* 09 MNH */ "#E0E0E0", 
	/* 10 SCH */ "#FF0000", /* 11 ALH */ "#008000",	/* 12 CTC */ "#FF0000", /* 13 HRM */ "#A0A0A0", 
	/* 14 MNM */ "#000000", /* 15 QUM */ "#C0C000", /* 16 GMF */ "#A0A000", /* 17 HHF */ "#606060", 
	/* 18 MNF */ "#A0A0A0", /* 19 CCF */ "#E00000", /* 20 SWH */ "#FF0000", /* 21 SWC */ "#FF0000", 
	/* 22 SWF */ "#E00000", /* 23 SWM */ "#6060FF"
#endif
};

GdkColor clockColours[MAX__COLOURS];

char *nameFormats[TXT_COUNT] =
{
	"ftl", "ftz", "stl", "stz", "fbl", "fbz", "sbl", "sbz", 
	"wtl", "wtz", "cbl", "cbz", "cdl", "cdz", "ctl", "ctz"
};

char displayString[TXT_COUNT][61] =
{
	"%a. %e %b.",					/* Clock top 		(localtime)		00 */
	"%#",							/* Clock top		(timezone) 		01 */
	"%e %b.%n%a.",					/* Clock SW top 	(localtime)		02 */
	"%#%n",							/* Clock SW top		(timezone) 		03 */
	"- %Z -",						/* Clock bottom		(localtime)		04 */
	"- %p -",						/* Clock bottom		(timezone) 		05 */
	"%Z",							/* Clock SW bottom	(localtime)		06 */
	"%p",							/* Clock SW bottom	(timezone) 		07 */	
	"%R, %A %e %B",					/* Window title 	(localtime)		08 */
	"%*, %R, %A %e %B",				/* Window title		(timezone) 		09 */
	"%A, %e %B %Y, %X",				/* Copy date time 	(localtime)		10 */
	"%A, %e %B %Y, %X, %*",			/* Copy date time	(timezone) 		11 */
	"%A, %e %B %Y",					/* Copy date 		(localtime)		12 */
	"%A, %e %B %Y, %*",				/* Copy date		(timezone) 		13 */
	"%X",							/* Copy time 		(localtime)		14 */
	"%X, %Z"						/* Copy time		(timezone) 		15 */
};

/*----------------------------------------------------------------------------------------------------*
 * If we cannot find a stock clock icon then use this built in one.                                   *
 *----------------------------------------------------------------------------------------------------*/
#include "TzClockIcon.xpm"

/*----------------------------------------------------------------------------------------------------*
 * Who dunit!                                                                                         *
 *----------------------------------------------------------------------------------------------------*/
const gchar *authors[] =
{
	"Chris Knight",
	"Eric Smith",
	NULL
};

const gchar *artists[] =
{
	"Annie Knight",
	"Sonya Knight",
	NULL
};

tzInfo timeZones[] = 
{
	#define TIMEZONE_TABLE
	#include "TimeZone.h"
	#undef TIMEZONE_TABLE	
};

static gint nTimeZones = sizeof (timeZones) / sizeof (tzInfo);

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
static void aboutCallback 			(GtkWidget * widget, guint data);
static void onTopCallback 			(GtkWidget * widget, guint data);
static void stickCallback 			(GtkWidget * widget, guint data);
static void lockCallback 			(GtkWidget * widget, guint data);
static void calendarCallback 		(GtkWidget * widget, guint data);
static void alarmCallback	 		(GtkWidget * widget, guint data);
static void setTimeZoneCallback 	(GtkWidget * widget, guint data);
static void copyCallback 			(GtkWidget * widget, guint data);
static void stopwatchCallback		(GtkWidget * widget, guint data);
static void subSecondCallback		(GtkWidget * widget, guint data);
static void showSecondsCallback		(GtkWidget * widget, guint data);
static void swStartCallback 		(GtkWidget * widget, guint data);
static void zoomCallback 			(GtkWidget * widget, guint data);
static void faceCallback			(GtkWidget * widget, guint data);
static void swResetCallback 		(GtkWidget * widget, guint data);
static void quitCallback 			(GtkWidget * widget, guint data);

/*----------------------------------------------------------------------------------------------------*
 * These need to match the menu items, they are used later to get and set the values                  *
 *----------------------------------------------------------------------------------------------------*/
#define PREF_ONTOP		"/Preferences/Always On Top"
#define PREF_DESKTOP	"/Preferences/All Desktops"
#define PREF_LOCK		"/Preferences/Lock Position"
#define PREF_SECS		"/Preferences/Show Seconds"
#define PREF_SUBS		"/Preferences/Sub Seconds"
#define SWTH_ENABLE		"/Stopwatch/Enabled"
#define SWTH_START		"/Stopwatch/Start+Stop"
#define SWTH_RESET		"/Stopwatch/Reset"
#define VIEW_ADDCOL		"/Preferences/View/Add Column"
#define VIEW_DELCOL		"/Preferences/View/Remove Column"
#define VIEW_ADDROW		"/Preferences/View/Add Row"
#define VIEW_DELROW		"/Preferences/View/Remove Row"

/*----------------------------------------------------------------------------------------------------*
 * Structure to store the menus, very closly linked to the timezones above.                           *
 *----------------------------------------------------------------------------------------------------*/
static GtkItemFactoryEntry menuItems[] = 
{ 
/*----------------------------------------------------------------------------------------------------*/
{ "/_Timezone", 					NULL,		NULL,				 0,	"<Branch>"					  	  },
/*----------------------------------------------------------------------------------------------------*/
#define TIMEZONE_MENU
#include "TimeZone.h"
#undef TIMEZONE_MENU
/*----------------------------------------------------------------------------------------------------*/
{ "/Stop_watch",				NULL,		NULL,				 0,	"<Branch>"					  	  }, 
{ "/Stopwatch/Ena_bled",		NULL,		stopwatchCallback,	 0,	"<ToggleItem>"				  	  },
{ "/Stopwatch/Start+St_op",		"<CTRL>S",	swStartCallback,	 0,	"<Item>"					  	  },
{ "/Stopwatch/_Reset",			"<CTRL>R",	swResetCallback,	 0,	"<Item>"					  	  },
/*----------------------------------------------------------------------------------------------------*/
{ "/_Edit",						NULL,		NULL,				 0,	"<Branch>"						  }, 
{ "/Edit/_Copy Date & Time",	"<CTRL>C",	copyCallback,		 0,	"<StockItem>", GTK_STOCK_COPY 	  },
{ "/Edit/Copy _Date",			"<CTRL>D",	copyCallback,		 1,	"<Item>"					  	  },
{ "/Edit/Copy _Time",			"<CTRL>T",	copyCallback,		 2,	"<Item>"					  	  },
/*----------------------------------------------------------------------------------------------------*/
{ "/SEP",	 					NULL,		NULL,				 0,	"<Separator>"				  	  },
/*----------------------------------------------------------------------------------------------------*/
{ "/A_larm Setup...",			NULL,		alarmCallback,	 	 0,	"<Item>"					  	  },
{ "/Show Cale_ndar...",			NULL,		calendarCallback,	 0,	"<Item>"					  	  },
/*----------------------------------------------------------------------------------------------------*/
{ "/SEP",	 					NULL,		NULL,				 0,	"<Separator>"				  	  },
/*----------------------------------------------------------------------------------------------------*/
{ "/_Preferences", 				NULL,		NULL, 				 0,	"<Branch>"					  	  },
{ "/Preferences/Always On _Top",NULL,		onTopCallback,		 0,	"<ToggleItem>"				  	  },
{ "/Preferences/All _Desktops",	NULL,		stickCallback,		 0,	"<ToggleItem>"					  },
{ "/Preferences/Lock _Position",NULL,		lockCallback,		 0,	"<ToggleItem>"					  },
{ "/Preferences/SEP3",			NULL,		NULL,				 0,	"<Separator>"					  },
{ "/Preferences/Show _Seconds",	NULL,		showSecondsCallback, 0,	"<ToggleItem>"					  },
{ "/Preferences/S_ub Seconds",	NULL,		subSecondCallback,	 0,	"<ToggleItem>"					  },
{ "/Preferences/SEP",	 		NULL,		NULL,				 0,	"<Separator>"				  	  },
/*----------------------------------------------------------------------------------------------------*/
{ "/Preferences/_Zoom",		 	NULL,		NULL, 				 0,	"<Branch>"						  },
{ "/Preferences/Zoom/_In",		"<CTRL>I",	zoomCallback,	 	 1,	"<StockItem>", GTK_STOCK_ZOOM_IN  },
{ "/Preferences/Zoom/_Out",		"<CTRL>O",	zoomCallback,	 	 2,	"<StockItem>", GTK_STOCK_ZOOM_OUT },
#ifdef GTK_STOCK_FULLSCREEN /* Support a stock fullscreen? */
{ "/Preferences/Zoom/Ma_x",		"<CTRL>X",	zoomCallback,	 	 3,	"<StockItem>", GTK_STOCK_FULLSCREEN },
#endif
/*----------------------------------------------------------------------------------------------------*/
{ "/Preferences/_View",		 			NULL,	NULL, 			 0,	"<Branch>"						  },
{ "/Preferences/View/_Add Column",		NULL,	faceCallback,	 1,	"<StockItem>", GTK_STOCK_ADD	  },
{ "/Preferences/View/_Remove Column",	NULL,	faceCallback,	 2,	"<StockItem>", GTK_STOCK_REMOVE	  },
{ "/Preferences/View/SEP", 				NULL,	NULL,			 0,	"<Separator>"					  },
{ "/Preferences/View/A_dd Row",			NULL,	faceCallback,	 3,	"<StockItem>", GTK_STOCK_ADD	  },
{ "/Preferences/View/R_emove Row",		NULL,	faceCallback,	 4,	"<StockItem>", GTK_STOCK_REMOVE	  },
/*----------------------------------------------------------------------------------------------------*/
{ "/SEP",	 					NULL,		NULL,				 0,	"<Separator>"					  },
#ifdef GTK_STOCK_ABOUT /* Support a stock About box? */
{ "/_About",					NULL,		aboutCallback,		 0,	"<StockItem>", GTK_STOCK_ABOUT	  },
#else
{ "/_About",					NULL,		aboutCallback,		 0,	"<StockItem>", GTK_STOCK_HELP 	  },
#endif
{ "/_Quit",						"<CTRL>Q",	quitCallback,		 0,	"<StockItem>", GTK_STOCK_QUIT	  }
/*----------------------------------------------------------------------------------------------------*/
};

static gint nMenuItems = sizeof (menuItems) / sizeof (GtkItemFactoryEntry);

/*----------------------------------------------------------------------------------------------------*
 * Use tables for the sin and cos calculation, it is faster.                                          *
 *----------------------------------------------------------------------------------------------------*/
static double sinTable[360];
static double cosTable[360];

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
static GtkWidget *menuBar;			/* actually a popup menu in transparency/shape mode */
static GtkItemFactory *itemFactory;
static GdkPixbuf *defaultIcon;
static GdkColormap *colourMap;
#ifdef STATUS_ICON
static GtkStatusIcon *statusIcon;
#endif

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
static int alwaysOnTop 			= 0;
static int stuckOnAll 			= 0;
static int lockMove 			= 0;
static int lastKeyPressTime		= 0;
static int keyPressFaceNum		= -1;
static int stopwatchActive		= 0;
static int alarmCounter			= 0;
static int useThemeIcon			= 0;
static time_t lastTime 			= -1;

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
GtkWidget *mainWindow;
GtkWidget *drawingArea;
int faceSize 			= 3 * 64;
int faceWidth 			= 1;
int faceHeight 			= 1;
int weHaveFocus 		= 0;
int currentFace 		= 0;
int timeSetting			= 0;
int fastSetting			= 0;
time_t forceTime		= -1;
FACE_SETTINGS faceSettings[MAX_FACES];

/*----------------------------------------------------------------------------------------------------*
 * Prototypes for functions in the tables that are defined later.                                     *
 *----------------------------------------------------------------------------------------------------*/
static void fillSinCosTables 		(void);
static void splitTimeZone 			(char *timeZone, char *area, char *city, char *display, int doUpper);
static GtkWidget *createMenus 		(GtkWidget * window, GtkType containerType);
static void createColours 			(void);
static void processCommandLine 		(int argc, char *argv[], int *posX, int *posY);
static void addStringValue 			(char *addBuffer, int maxSize, char *stringFormat, int face, time_t timeNow);
static void howTo 					(FILE * outFile, char *format, ...);
static void fixFaceSize				(void);

static gboolean clockTickCallback	(gpointer data);
static gboolean windowClickCallback	(GtkWidget * widget, GdkEventButton * event);
static gboolean windowKeyCallback	(GtkWidget * widget, GdkEventKey * event);
static gboolean exposeCallback 		(GtkWidget * widget, GdkEventExpose * event, gpointer data);
static gboolean focusInEvent 		(GtkWidget *widget, GdkEventFocus *event, gpointer data);
static gboolean focusOutEvent 		(GtkWidget *widget, GdkEventFocus *event, gpointer data);
static gboolean statusActivate		(GtkWidget *widget, gpointer data);

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  H O W  T O                                                                                        *
 *  ==========                                                                                        *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param outFile .
 *  @param format .
 *  @param ... .
 *  @result .
 */
void
howTo (FILE * outFile, char *format, ...)
{
	va_list ap;

	fprintf (outFile, "------------------------------------------------------------\n");
	fprintf (outFile, "The Clock %s\n", VERSION);
	fprintf (outFile, "------------------------------------------------------------\n");
	fprintf (outFile, "How to use: TzClock [options...]\n\n");
	fprintf (outFile, "   -a           :  Always on top\n");
	fprintf (outFile, "   -Ahh:mm:msg  :* Alarm message shown at a specified time\n");
	fprintf (outFile, "   -b           :* Small second hand\n");
	fprintf (outFile, "   -cnnn#RRGGBB :  Change one of the clock colours\n");
	fprintf (outFile, "   -dnnn:format :  Change one of the date formats\n");
	fprintf (outFile, "   -f<face>     :  Select the face, for setting timezone\n");
	fprintf (outFile, "   -F           :  Fast time setting, no smooth scroll\n");
	fprintf (outFile, "   -h           :* Switch off the second hand\n");
	fprintf (outFile, "   -i           :  Use the clock icon from the current theme\n");
	fprintf (outFile, "   -l           :  Lock position\n");
	fprintf (outFile, "   -n<c|r><num> :  Set the number of columns and rows\n");
	fprintf (outFile, "                :  Max clocks %d, no more than 10 in a line\n", MAX_FACES);
	fprintf (outFile, "   -s<size>     :  Set the size of each clock\n");
	fprintf (outFile, "   -S           :* Enable the stopwatch\n");
	fprintf (outFile, "   -u           :* Uppercase the city name\n");
	fprintf (outFile, "   -w           :  Put on all workspaces\n");
	fprintf (outFile, "   -x<posn>     :  Set the X screen position\n");
	fprintf (outFile, "   -y<posn>     :  Set the Y screen position\n");
	fprintf (outFile, "                :  Both X and Y must be set\n");
	fprintf (outFile, "   -z<zone>     :* Select timezone to display\n");
	fprintf (outFile, "   -24          :* Show a 24 hour clock\n");
	fprintf (outFile, "   -?           :  This how to information\n\n");
	fprintf (outFile, "Options marked with '*' only effect the current face. Use\n");
	fprintf (outFile, "the -f<num> option to select the current face.\n");
	fprintf (outFile, "------------------------------------------------------------\n");
	fprintf (outFile, "Colour codes: -cnnn#RRGGBB  (nnn Colour name)\n\n");

	fprintf (outFile, "   %s : Clock border (focused)\n", 			nameColours[2]);
	fprintf (outFile, "   %s : Clock border (normal)\n", 			nameColours[3]);
	fprintf (outFile, "   %s : Clock border (inner)\n",				nameColours[4]);
	fprintf (outFile, "   %s : Clock face colour\n",	 			nameColours[5]);
	fprintf (outFile, "   %s : Stopwatch face colour\n",			nameColours[6]);	
	fprintf (outFile, "   %s : Text colour\n",						nameColours[7]);
	fprintf (outFile, "   %s : Hour hand\n", 						nameColours[8]);
	fprintf (outFile, "   %s : Minute hand\n", 						nameColours[9]);
	fprintf (outFile, "   %s : Second hand\n", 						nameColours[10]);
	fprintf (outFile, "   %s : Alarm hand\n",						nameColours[11]);	
	fprintf (outFile, "   %s : Center circle\n", 					nameColours[12]);
	fprintf (outFile, "   %s : Hour markers\n", 					nameColours[13]);
	fprintf (outFile, "   %s : Minute markers\n", 					nameColours[14]);
	fprintf (outFile, "   %s : Quarter markers\n", 					nameColours[15]);
	fprintf (outFile, "   %s : Quarter marker fill\n", 				nameColours[16]);
	fprintf (outFile, "   %s : Hour hand fill\n", 					nameColours[17]);
	fprintf (outFile, "   %s : Minute hand fill\n", 				nameColours[18]);
	fprintf (outFile, "   %s : Center circle fill\n", 				nameColours[19]);
	fprintf (outFile, "   %s : Stopwatch second hand\n",			nameColours[20]);
	fprintf (outFile, "   %s : Stopwatch center circle\n",			nameColours[21]);
	fprintf (outFile, "   %s : Stopwatch circle fill\n",			nameColours[22]);
	fprintf (outFile, "   %s : Stopwatch minute markers\n",			nameColours[23]);
	fprintf (outFile, "------------------------------------------------------------\n");
	fprintf (outFile, "Date format strings: -dnnn:format (nnn Format name)\n\n");

	fprintf (outFile, "   %s : Clock top in localtime\n", 			nameFormats[0]);
	fprintf (outFile, "   %s : Clock top for timezone\n", 			nameFormats[1]);
	fprintf (outFile, "   %s : Stopwatch top in localtime\n", 		nameFormats[2]);
	fprintf (outFile, "   %s : Stopwatch top for timezone\n", 		nameFormats[3]);
	fprintf (outFile, "   %s : Clock bottom in localtime\n", 		nameFormats[4]);
	fprintf (outFile, "   %s : Clock bottom for timezone\n", 		nameFormats[5]);
	fprintf (outFile, "   %s : Stopwatch bottom in localtime\n", 	nameFormats[6]);
	fprintf (outFile, "   %s : Stopwatch bottom for timezone\n", 	nameFormats[7]);
	fprintf (outFile, "   %s : Window title in localtime\n", 		nameFormats[8]);
	fprintf (outFile, "   %s : Window title for timezone\n", 		nameFormats[9]);
	fprintf (outFile, "   %s : Copy date time in localtime\n", 		nameFormats[10]);
	fprintf (outFile, "   %s : Copy date time for timezone\n", 		nameFormats[11]);
	fprintf (outFile, "   %s : Copy date in localtime\n", 			nameFormats[12]);
	fprintf (outFile, "   %s : Copy date for timezone\n", 			nameFormats[13]);
	fprintf (outFile, "   %s : Copy time in localtime\n", 			nameFormats[14]);
	fprintf (outFile, "   %s : Copy time for timezone\n", 			nameFormats[15]);
	fprintf (outFile, "------------------------------------------------------------\n");
	fprintf (outFile, "Date format options:\n\n");

	fprintf (outFile, "   %%#   : Time zone city, uppercased and wrapped\n");
	fprintf (outFile, "   %%*   : Time zone city\n");
	fprintf (outFile, "   %%@   : Time zone area\n");
	fprintf (outFile, "   %%&   : Stopwatch time: h:mm:ss.hh\n");
	fprintf (outFile, "   %%... : See man page for the date command\n");

	fprintf (outFile, "------------------------------------------------------------\n");
	if (format)
	{
		va_start(ap, format);
		vfprintf(outFile, format, ap);
		va_end(ap);
		fprintf (outFile, "------------------------------------------------------------\n");
	}
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  S P L I T  T I M E  Z O N E                                                                       *
 *  ===========================                                                                       *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param timeZone .
 *  @param area .
 *  @param city .
 *  @param display .
 *  @param doUpper .
 *  @result .
 */
static void
splitTimeZone (char *timeZone, char *area, char *city, char *display, int doUpper)
{
	int i = 0, j = 0, w = 0, newLine = 0, wordSize = 0, addCR = 0;
	char inChar;

	if (area)
		area[j] = 0;

	while (timeZone[i])
	{
		inChar = timeZone[i++];

		if (inChar == '_')
		{
			if (j >= 5 && newLine == 0 && w == 1 && wordSize >= 12)
			{
				addCR = 1;
				newLine = 1;
			}
			inChar = ' ';
		}
		if (inChar == '/')
		{
			w = 1;
			j = 0;

			wordSize = strlen (&timeZone[i]);
			if (city)
				city [j] = 0;
			if (display)
				display[j] = 0;
		}
		else if (w == 0)
		{
			if (area)
			{
				area[j++] = inChar;
				area[j] = 0;
			}
		}
		else
		{
			if (city)
			{
				city[j] = inChar;
				city[j + 1] = 0;
			}
			if (display)
			{
				char s = addCR ? '\n' : inChar;
				display[j] = doUpper ? toupper (s) : s;
				display[j + 1] = 0;
			}
			j++;
		}
		addCR = 0;
	}
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  Q U I T  C A L L B A C K                                                                          *
 *  ========================                                                                          *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param data .
 *  @result .
 */
static void
quitCallback (GtkWidget * widget, guint data)
{
	gtk_main_quit ();
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  C O P Y  C A L L B A C K                                                                          *
 *  ========================                                                                          *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param data .
 *  @result .
 */
static void
copyCallback (GtkWidget * widget, guint data)
{
	time_t t;
	char stringDate[101];
	GtkClipboard* clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
	int timeZone = faceSettings[currentFace].currentTZ;

	t = time (NULL);
	if (timeZones[timeZone].value == 0)
	{
		unsetenv ("TZ");
	}
	else
	{
		setenv ("TZ", timeZones[timeZone].envName, 1);
	}
	localtime (&t);

	switch (data)
	{
	case 0:
		getStringValue (stringDate, NULL, 100, timeZone ? TXT_COPY_DT_Z : TXT_COPY_DT_L, currentFace, t);
		break;

	case 1:
		getStringValue (stringDate, NULL, 100, timeZone ? TXT_COPY_D_Z : TXT_COPY_D_L, currentFace, t);
		break;

	case 2:
		getStringValue (stringDate, NULL, 100, timeZone ? TXT_COPY_T_Z : TXT_COPY_T_L, currentFace, t);
		break;
	}
	gtk_clipboard_set_text (clipboard, stringDate, -1);
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  S E T  T I M E  Z O N E  C A L L B A C K                                                          *
 *  ========================================                                                          *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param data .
 *  @result .
 */
static void
setTimeZoneCallback (GtkWidget * widget, guint data)
{
	int timeZone = (int) data;
	
	faceSettings[currentFace].currentTZ = timeZones[timeZone].value;
	splitTimeZone (timeZones[timeZone].envName, faceSettings[currentFace].currentTZArea, 
			faceSettings[currentFace].currentTZCity, faceSettings[currentFace].currentTZDisp,
			faceSettings[currentFace].upperCity);

	lastTime = -1;
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  O N  T O P  C A L L B A C K                                                                       *
 *  ===========================                                                                       *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param data .
 *  @result .
 */
static void
onTopCallback (GtkWidget * widget, guint data)
{
	alwaysOnTop = gtk_check_menu_item_get_active(
			GTK_CHECK_MENU_ITEM (gtk_item_factory_get_item (itemFactory, PREF_ONTOP)));
	gtk_window_set_keep_above (GTK_WINDOW (mainWindow), alwaysOnTop != 0);
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  S T I C K  C A L L B A C K                                                                        *
 *  ==========================                                                                        *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param data .
 *  @result .
 */
static void stickCallback (GtkWidget * widget, guint data)
{
	stuckOnAll = gtk_check_menu_item_get_active(
			GTK_CHECK_MENU_ITEM (gtk_item_factory_get_item (itemFactory, PREF_DESKTOP)));

	if (stuckOnAll)
	{
		gtk_window_stick (GTK_WINDOW (mainWindow));
	}
	else
	{
		gtk_window_unstick (GTK_WINDOW (mainWindow));
	}
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  L O C K  C A L L B A C K                                                                          *
 *  ========================                                                                          *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param data .
 *  @result .
 */
static void
lockCallback (GtkWidget * widget, guint data)
{
	lockMove = gtk_check_menu_item_get_active(
			GTK_CHECK_MENU_ITEM (gtk_item_factory_get_item (itemFactory, PREF_LOCK)));
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  C A L E N D A R  C A L L B A C K                                                                  *
 *  ================================                                                                  *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param data .
 *  @result .
 */
static void
calendarCallback (GtkWidget * widget, guint data)
{
	time_t t;
	GtkWidget *dialog;
	int timeZone = faceSettings[currentFace].currentTZ;

	t = time (NULL);
	if (timeZones[timeZone].value == 0)
	{
		unsetenv ("TZ");
	}
	else
	{
		setenv ("TZ", timeZones[timeZone].envName, 1);
	}
	localtime (&t);
	
	dialog = gtk_dialog_new_with_buttons ("TzClock Calendar", GTK_WINDOW (mainWindow),
					      GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT, GTK_STOCK_OK,
					      GTK_RESPONSE_NONE, NULL);

	gtk_dialog_set_has_separator (GTK_DIALOG (dialog), TRUE);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), gtk_calendar_new ());
	gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_CENTER);

	gtk_widget_show_all (dialog);
	gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_widget_destroy (dialog);
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  A L A R M  C A L L B A C K                                                                        *
 *  ==========================                                                                        *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param data .
 *  @result .
 */
static void
alarmCallback (GtkWidget * widget, guint data)
{
	GtkWidget *dialog;
	GtkWidget *frame;
    GtkWidget *entry;
	GtkWidget *label;
	GtkWidget *spinner1, *spinner2;
	GtkWidget *hbox, *vbox1, *vbox2;
	GtkAdjustment *adj;
	
	/*------------------------------------------------------------------------------------------------*
	 * Create the basic dialog box                                                                    *
	 *------------------------------------------------------------------------------------------------*/	
	dialog = gtk_dialog_new_with_buttons ("Alarm", GTK_WINDOW (mainWindow),
			GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT, 
			GTK_STOCK_CANCEL,
			GTK_RESPONSE_REJECT,
			GTK_STOCK_SAVE,
			GTK_RESPONSE_ACCEPT, 
			NULL);

	gtk_dialog_set_has_separator (GTK_DIALOG(dialog), false);
	
	frame = gtk_frame_new ("Setup");
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG(dialog)->vbox), frame, TRUE, TRUE, 0);
	
	vbox1 = gtk_vbox_new (FALSE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (vbox1), 7);
	gtk_container_add (GTK_CONTAINER (frame), vbox1);

	hbox = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox1), hbox, TRUE, TRUE, 5);
	
	/*------------------------------------------------------------------------------------------------*
	 * Add the hour spinner                                                                           *
	 *------------------------------------------------------------------------------------------------*/	
	vbox2 = gtk_vbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX (hbox), vbox2, FALSE, TRUE, 0);

	label = gtk_label_new ("Hour :");
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_box_pack_start (GTK_BOX (vbox2), label, FALSE, TRUE, 0);

	adj = (GtkAdjustment *) gtk_adjustment_new 
			(faceSettings[currentFace].alarmInfo.alarmHour, 0, 23, 1, 4, 0);
	spinner1 = gtk_spin_button_new (adj, 0, 0);
	gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (spinner1), true);
	gtk_spin_button_set_snap_to_ticks (GTK_SPIN_BUTTON (spinner1), true);
	gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (spinner1), TRUE);
	gtk_box_pack_start (GTK_BOX (vbox2), spinner1, FALSE, TRUE, 0);

	/*------------------------------------------------------------------------------------------------*
	 * Add the minute spinner                                                                         *
	 *------------------------------------------------------------------------------------------------*/	
	vbox2 = gtk_vbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX (hbox), vbox2, FALSE, TRUE, 5);

	label = gtk_label_new ("Min :");
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_box_pack_start (GTK_BOX (vbox2), label, FALSE, TRUE, 0);

	adj = (GtkAdjustment *) gtk_adjustment_new
			(faceSettings[currentFace].alarmInfo.alarmMin, 0, 59, 1, 5, 0);
	spinner2 = gtk_spin_button_new (adj, 0, 0);
	gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (spinner2), true);
	gtk_spin_button_set_snap_to_ticks (GTK_SPIN_BUTTON (spinner2), true);
	gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (spinner2), TRUE);
	gtk_box_pack_start (GTK_BOX (vbox2), spinner2, FALSE, TRUE, 0);

	/*------------------------------------------------------------------------------------------------*
	 * Add the message entry                                                                          *
	 *------------------------------------------------------------------------------------------------*/	
	vbox2 = gtk_vbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox1), vbox2, TRUE, TRUE, 5);

	label = gtk_label_new ("Message :");
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_box_pack_start (GTK_BOX(vbox2), label, FALSE, TRUE, 0);

	entry = gtk_entry_new ();
	gtk_entry_set_max_length (GTK_ENTRY (entry), 40);
	gtk_entry_set_text (GTK_ENTRY (entry), faceSettings[currentFace].alarmInfo.message);
	gtk_editable_select_region (GTK_EDITABLE (entry), 0, GTK_ENTRY (entry)->text_length);
    gtk_box_pack_start (GTK_BOX(vbox2), entry, TRUE, TRUE, 0);

	/*------------------------------------------------------------------------------------------------*
	 * Display it, if OK pressed the save the new values                                              *
	 *------------------------------------------------------------------------------------------------*/	
	gtk_widget_show_all (dialog);
	if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
	{
		strcpy (faceSettings[currentFace].alarmInfo.message, gtk_entry_get_text (GTK_ENTRY(entry)));
		faceSettings[currentFace].alarmInfo.alarmHour = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON(spinner1));
		faceSettings[currentFace].alarmInfo.alarmMin  = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON(spinner2));
		faceSettings[currentFace].alarmInfo.showAlarm = (faceSettings[currentFace].alarmInfo.message[0] != 0);
    }	
	gtk_widget_destroy (dialog);
}

#ifdef GTK_ABOUT_DIALOG

static void
show_url (GtkWidget *parent, const char *url)
{
	GdkScreen *screen;
	GError *error = NULL;

	screen = gtk_widget_get_screen (parent);

	if (!gnome_url_show_on_screen (url, screen, &error))
	{
		GtkWidget *dialog;

		dialog = gtk_message_dialog_new (GTK_WINDOW (parent),
				GTK_DIALOG_DESTROY_WITH_PARENT,
				GTK_MESSAGE_ERROR,
				GTK_BUTTONS_OK,
				"%s", "Could not open link");
		gtk_message_dialog_format_secondary_text
				(GTK_MESSAGE_DIALOG (dialog), "%s", error->message);
		g_error_free (error);

		g_signal_connect (dialog, "response",
		G_CALLBACK (gtk_widget_destroy), NULL);
		gtk_widget_show (dialog);
	}
}

void aboutUrlHook (GtkAboutDialog *about, const gchar *link_, gpointer data)
{
	show_url (GTK_WIDGET (about), link_);
}

#endif

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  A B O U T  C A L L B A C K                                                                        *
 *  ==========================                                                                        *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param data .
 *  @result .
 */
static void
aboutCallback (GtkWidget * widget, guint data)
{
	GtkWidget *aboutDialog;
	
	char comment[256];
	char *webSite = "http://www.TzClock.org/";
	char *copyright = "(c) Copyright 2005 - 2007";
	
#ifdef BUILDCAIRO
	sprintf (comment, "TzClock is a highly configurable Cairo based analog clock,\n"
					  "capable of showing the time in different timezones.\n"
					  "Cairo version: %s", cairo_version_string ());
#else
	sprintf (comment, "TzClock is a highly configurable GTK based analog clock,\n"
					  "capable of showing the time in different timezones.\n"
					  "GTK version: %d.%d.%d", gtk_major_version, gtk_minor_version, gtk_micro_version);
#endif

#ifdef GTK_ABOUT_DIALOG

	/*------------------------------------------------------------------------------------------------*
	 * Nice dialog that can be used with newer versions of the GTK API.                               *
	 *------------------------------------------------------------------------------------------------*/
	aboutDialog = gtk_about_dialog_new ();
	gtk_about_dialog_set_url_hook (aboutUrlHook, NULL, NULL);

	gtk_about_dialog_set_version (GTK_ABOUT_DIALOG (aboutDialog), VERSION);
	gtk_about_dialog_set_copyright (GTK_ABOUT_DIALOG (aboutDialog), copyright);
	gtk_about_dialog_set_authors (GTK_ABOUT_DIALOG (aboutDialog), authors);
	gtk_about_dialog_set_artists (GTK_ABOUT_DIALOG (aboutDialog), artists);
	gtk_about_dialog_set_comments (GTK_ABOUT_DIALOG (aboutDialog), comment);
	gtk_about_dialog_set_website (GTK_ABOUT_DIALOG (aboutDialog), webSite);
	gtk_about_dialog_set_logo (GTK_ABOUT_DIALOG (aboutDialog), defaultIcon);

#else

	/*------------------------------------------------------------------------------------------------*
	 * Older version of the about dialog that has to be used with older versions of the API.          *
	 *------------------------------------------------------------------------------------------------*/
	char buf[80];
	
	aboutDialog = gtk_dialog_new_with_buttons ("About TzClock", GTK_WINDOW (mainWindow),
			GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT, GTK_STOCK_OK,
			GTK_RESPONSE_NONE, NULL);
			
	gtk_dialog_set_has_separator (GTK_DIALOG (aboutDialog), TRUE);
	sprintf (buf, "*        TzClock Release: %s        *\n", VERSION);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (aboutDialog)->vbox), gtk_label_new (buf));
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (aboutDialog)->vbox), gtk_label_new (comment));
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (aboutDialog)->vbox), gtk_label_new (copyright));
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (aboutDialog)->vbox), gtk_label_new (webSite));
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (aboutDialog)->vbox), gtk_label_new ("Authors:"));
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (aboutDialog)->vbox), gtk_label_new ("Chris Knight"));
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (aboutDialog)->vbox), gtk_label_new ("Eric Smith"));
	
	gtk_widget_show_all (aboutDialog);

#endif /* GTK_ABOUT_DIALOG */

	gtk_dialog_run (GTK_DIALOG (aboutDialog));
	gtk_widget_destroy (aboutDialog);
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  C R E A T E  M E N U S                                                                            *
 *  ======================                                                                            *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param window .
 *  @param containerType .
 *  @result .
 */
static GtkWidget *
createMenus (GtkWidget * window, GtkType containerType)
{
	GtkAccelGroup *accelGroup;

	accelGroup = gtk_accel_group_new ();
	itemFactory = gtk_item_factory_new (containerType, "<main>", accelGroup);
	gtk_item_factory_create_items (itemFactory, nMenuItems, menuItems, NULL);
	gtk_window_add_accel_group (GTK_WINDOW (window), accelGroup);

	return (gtk_item_factory_get_widget (itemFactory, "<main>"));
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  C R E A T E  C O L O U R S                                                                        *
 *  ==========================                                                                        *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @result .
 */
void
createColours (void)
{
	int i;
	char *setColour;
	colourMap = gdk_colormap_get_system ();

	for (i = 0; i < MAX__COLOURS; i++)
	{
		setColour = (delaultColour[i]);
		gdk_color_parse (setColour, &clockColours[i]);
		gdk_colormap_alloc_color (colourMap, &clockColours[i], FALSE, FALSE);
	}
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  P R E P A R E  F O R  P O P U P                                                                   *
 *  ===============================                                                                   *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @result .
 */
void prepareForPopup (void)
{
	/*------------------------------------------------------------------------------------------------* 
	 * The start stop options are only avalable if the current face has the stopwatch enabled         * 
	 *------------------------------------------------------------------------------------------------*/
	gtk_widget_set_sensitive (
			GTK_WIDGET (gtk_item_factory_get_item (itemFactory, SWTH_START)),
			faceSettings[currentFace].stopwatch ? TRUE : FALSE);

	gtk_widget_set_sensitive (
			GTK_WIDGET (gtk_item_factory_get_item (itemFactory, SWTH_RESET)),
			(faceSettings[currentFace].stopwatch && (faceSettings[currentFace].swStartTime != -1 ||
			faceSettings[currentFace].swRunTime != 0))
					 ? TRUE : FALSE);
			
	/*------------------------------------------------------------------------------------------------* 
	 * The sub second option is only available if seconds are being show                              * 
	 *------------------------------------------------------------------------------------------------*/
	gtk_widget_set_sensitive (
			GTK_WIDGET (gtk_item_factory_get_item (itemFactory, PREF_SUBS)),
			faceSettings[currentFace].showSeconds ? TRUE : FALSE);

	/*------------------------------------------------------------------------------------------------* 
	 * Set the active state of the option to match the current setting on the face                    * 
	 *------------------------------------------------------------------------------------------------*/
	gtk_check_menu_item_set_active(
			GTK_CHECK_MENU_ITEM (gtk_item_factory_get_item (itemFactory, SWTH_ENABLE)),	
			faceSettings[currentFace].stopwatch ? TRUE : FALSE);

	gtk_check_menu_item_set_active(
			GTK_CHECK_MENU_ITEM (gtk_item_factory_get_item (itemFactory, PREF_SECS)),
			faceSettings[currentFace].showSeconds ? TRUE : FALSE);

	gtk_check_menu_item_set_active(
			GTK_CHECK_MENU_ITEM (gtk_item_factory_get_item (itemFactory, PREF_SUBS)),
			faceSettings[currentFace].subSecond ? TRUE : FALSE);
			
	/*------------------------------------------------------------------------------------------------* 
	 *                                                                                                * 
	 *------------------------------------------------------------------------------------------------*/
	gtk_widget_set_sensitive(
			GTK_WIDGET (gtk_item_factory_get_item (itemFactory, VIEW_ADDCOL)),
			(faceWidth < 10 && ((faceWidth + 1) * faceHeight) <= MAX_FACES) ? TRUE : FALSE);

	gtk_widget_set_sensitive(
			GTK_WIDGET (gtk_item_factory_get_item (itemFactory, VIEW_DELCOL)),
			(faceWidth > 1) ? TRUE : FALSE);

	gtk_widget_set_sensitive(
			GTK_WIDGET (gtk_item_factory_get_item (itemFactory, VIEW_ADDROW)),
			(faceHeight < 10 && ((faceHeight + 1) * faceWidth) <= MAX_FACES) ? TRUE : FALSE);

	gtk_widget_set_sensitive(
			GTK_WIDGET (gtk_item_factory_get_item (itemFactory, VIEW_DELROW)),
			(faceHeight > 1) ? TRUE : FALSE);
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  W I N D O W  C L I C K  C A L L B A C K                                                           *
 *  =======================================                                                           *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param event .
 *  @result .
 */
gboolean
windowClickCallback (GtkWidget * widget, GdkEventButton * event)
{
	if (event->type == GDK_BUTTON_PRESS)
	{
		currentFace = ((int)event -> x / faceSize) + (((int)event -> y / faceSize) * faceWidth);
		lastTime = -1;

		switch (event->button)
		{
		case 1:	/* left button */
			if (!lockMove)
			{
				gtk_window_begin_move_drag (GTK_WINDOW (mainWindow), event->button, event->x_root,
						event->y_root, event->time);
			}
			return TRUE;
			
		case 3:	/* right button */
			prepareForPopup ();

			gtk_menu_popup (GTK_MENU (menuBar), NULL,	/* parent_menu_shell */
					NULL,								/* parent_menu_item */
					NULL,								/* func */
					NULL,								/* data */
					event->button, event->time);
			return TRUE;
		}
	}
	return FALSE;
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  W I N D O W  K E Y  C A L L B A C K                                                               *
 *  ===================================                                                               *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param event .
 *  @result .
 */
gboolean
windowKeyCallback (GtkWidget * widget, GdkEventKey * event)
{
	/*------------------------------------------------------------------------------------------------*
	 * You can select a face buy pressing and holding the ALT key while typing the number of the      *
	 * clock facce that you want to select.  In order to do this we track the ALT key press and       *
	 * release.  Each key must be pressed with in two seconds of the last key press.                  *
	 *------------------------------------------------------------------------------------------------*/
	if (event->time < lastKeyPressTime + 2000)
	{
		lastKeyPressTime = event->time;
	}
	else
	{
		lastKeyPressTime = 0;
		keyPressFaceNum = -1;
	}
	
	/*------------------------------------------------------------------------------------------------*
	 * Process key press events                                                                       *
	 *------------------------------------------------------------------------------------------------*/
	if (event->type == GDK_KEY_PRESS)
	{
		if (event->keyval == 0xFFE9) /* Alt key has been pressed */
		{
			lastKeyPressTime = event->time;
			keyPressFaceNum = 0;
		}
			
		if (event->state & 8) /* Trial and error showed 8 to be the Alt key flag */
		{
			/*----------------------------------------------------------------------------------------*
			 * Press ALT + M to display the Menu at the current mouse position                        *
			 *----------------------------------------------------------------------------------------*/
			if (event->keyval == GDK_M || event->keyval == GDK_m)
			{
				prepareForPopup ();
			
				gtk_menu_popup (GTK_MENU (menuBar), NULL,	/* parent_menu_shell */
						NULL,								/* parent_menu_item */
						NULL,								/* func */
						NULL,								/* data */
						0, event->time);
			}
			/*----------------------------------------------------------------------------------------*
			 * Press ALT + n where n is a number between 1 and 9 to select the face                   *
			 *----------------------------------------------------------------------------------------*/
			if (keyPressFaceNum != -1)
			{
				if (event->keyval >= GDK_0 && event->keyval <= GDK_9)
				{
					keyPressFaceNum *= 10;
					keyPressFaceNum += event->keyval - GDK_0;
					lastKeyPressTime = event->time;
				}
				else
				{
					keyPressFaceNum = -1;
					lastKeyPressTime = 0;
				}
			}
		}
		gtk_window_activate_key (GTK_WINDOW (mainWindow), event);
	}
	/*------------------------------------------------------------------------------------------------*
	 * Process key release events                                                                     *
	 *------------------------------------------------------------------------------------------------*/
	else if (event->type == GDK_KEY_RELEASE)
	{
		if (event->keyval == 0xFFE9 && keyPressFaceNum != -1)  /* Alt key released */
		{
			keyPressFaceNum --;
			if (keyPressFaceNum >= 0 && keyPressFaceNum < (faceWidth * faceHeight))
			{
				currentFace = keyPressFaceNum;
				lastTime = -1;
			}
		}
	}
	return TRUE;
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  C L O C K  T I C K  C A L L B A C K                                                               *
 *  ===================================                                                               *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param data .
 *  @result .
 */
gboolean
clockTickCallback (gpointer data)
{
	time_t timeNow = time (NULL);
	
	if (timeNow != lastTime || stopwatchActive || timeSetting)
	{
		GdkRectangle rect = 
		{
			0, 0, 0, 0
		};
		
		rect.width = drawingArea->allocation.width;
		rect.height = drawingArea->allocation.height;
		gdk_window_invalidate_rect (drawingArea->window, &rect, FALSE);
		lastTime = timeNow;
	}
	return TRUE;
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  F I L L  S I N  C O S  T A B L E S                                                                *
 *  ==================================                                                                *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @result .
 */
void
fillSinCosTables (void)
{
	int i;
	
	for (i = 0; i < 360; i++)
	{
		sinTable[i] = sin (((double) i * M_PI) / 180);
		cosTable[i] = cos (((double) i * M_PI) / 180);
	}
	
	/*------------------------------------------------------------------------------------------------*
	 * Extra table fix to remove rounding errors on the 90 degree angles.                             *
	 *------------------------------------------------------------------------------------------------*/
	sinTable[0]   =  0;
	cosTable[0]   =  1;
	sinTable[90]  =  1;
	cosTable[90]  =  0;
	sinTable[180] =  0;
	cosTable[180] = -1;
	sinTable[270] = -1;
	cosTable[270] =  0;
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  F O C U S  I N  E V E N T                                                                         *
 *  =========================                                                                         *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param event .
 *  @param data .
 *  @result .
 */
gboolean 
focusInEvent (GtkWidget *widget, GdkEventFocus *event, gpointer data)
{
	lastTime = -1;
	weHaveFocus = 1;
	return TRUE;
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  F O C U S  O U T  E V E N T                                                                       *
 *  ===========================                                                                       *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param event .
 *  @param data .
 *  @result .
 */
gboolean 
focusOutEvent (GtkWidget *widget, GdkEventFocus *event, gpointer data)
{
	lastTime = -1;
	weHaveFocus = 0;
	return TRUE;
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  S T A T U S  A C T I V A T E                                                                      *
 *  ============================                                                                      *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param data .
 *  @result .
 */
static gboolean statusActivate (GtkWidget *widget, gpointer data)
{
	gtk_window_present (GTK_WINDOW (mainWindow));
	return TRUE;
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  F I X  F A C E  S I Z E                                                                           *
 *  =======================                                                                           *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @result .
 */
void fixFaceSize (void)
{
	int max = 1024;
	
	if (max * faceWidth > gdk_screen_width())
	{
		max = ((gdk_screen_width() / faceWidth) * 64) / 64;
	}
	if (max * faceHeight > gdk_screen_height())
	{
		max = ((gdk_screen_height() / faceHeight) * 64) / 64;
	}
	if (faceSize > max)
	{
		faceSize = max;
	}
	else if (faceSize < 64)
	{
		faceSize = 64;
	}
	else
	{
		faceSize = ((faceSize + 32) / 64) * 64;
	}
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  Z O O M  C A L L B A C K                                                                          *
 *  ========================                                                                          *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param data .
 *  @result .
 */
static void
zoomCallback (GtkWidget * widget, guint data)
{
	switch (data)
	{
	case 1:
		faceSize += 64;
		break;
	case 2:
		faceSize -= 64;
		break;
	case 3:
		faceSize = 1024;
		break;
	}	
	fixFaceSize ();
	makeWindowMask();
	gtk_widget_set_size_request (drawingArea, faceWidth * faceSize, faceHeight * faceSize);
}	

static void
faceCallback (GtkWidget * widget, guint data)
{
	switch (data)
	{
	case 1:
		if (faceWidth < 10 && ((faceWidth + 1) * faceHeight) <= MAX_FACES)
			faceWidth ++;
		break;
	case 2:
		if (faceWidth > 1)
			faceWidth --;
		break;
	case 3:
		if (faceHeight < 10 && ((faceHeight + 1) * faceWidth) <= MAX_FACES)
			faceHeight ++;
		break;
	case 4:
		if (faceHeight > 1)
			faceHeight --;
		break;
	}	
	makeWindowMask();
	gtk_widget_set_size_request (drawingArea, faceWidth * faceSize, faceHeight * faceSize);
}	

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  S H O W  S E C O N D S  C A L L B A C K                                                           *
 *  =======================================                                                           *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param data .
 *  @result .
 */
static void
showSecondsCallback (GtkWidget * widget, guint data)
{
	faceSettings[currentFace].showSeconds = gtk_check_menu_item_get_active(
			GTK_CHECK_MENU_ITEM (gtk_item_factory_get_item (itemFactory, PREF_SECS)));
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  S U B  S E C O N D  C A L L B A C K                                                               *
 *  ===================================                                                               *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param data .
 *  @result .
 */
static void
subSecondCallback (GtkWidget * widget, guint data)
{
	faceSettings[currentFace].subSecond = gtk_check_menu_item_get_active(
			GTK_CHECK_MENU_ITEM (gtk_item_factory_get_item (itemFactory, PREF_SUBS)));
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  S T O P W A T C H  C A L L B A C K                                                                *
 *  ==================================                                                                *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param data .
 *  @result .
 */
static void
stopwatchCallback (GtkWidget * widget, guint data)
{
	int enabled;
	
	enabled = faceSettings[currentFace].stopwatch = 
			gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM (
			gtk_item_factory_get_item (itemFactory, SWTH_ENABLE)));
	
	faceSettings[currentFace].swStartTime = -1;
	faceSettings[currentFace].swRunTime = 0;
	
	gtk_widget_set_sensitive (
			GTK_WIDGET (gtk_item_factory_get_item (itemFactory, SWTH_START)), 
			enabled);

	gtk_widget_set_sensitive (
			GTK_WIDGET (gtk_item_factory_get_item (itemFactory, SWTH_RESET)), 
			enabled);
	
	lastTime = -1;
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  S W  S T A R T  C A L L B A C K                                                                   *
 *  ===============================                                                                   *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param data .
 *  @result .
 */
static void
swStartCallback (GtkWidget * widget, guint data)
{
	if (faceSettings[currentFace].stopwatch)
	{
		struct timeval tv;
		struct timezone tz;
	
		if (faceSettings[currentFace].swStartTime == -1)
		{
			if (gettimeofday(&tv, &tz) == 0)
			{
				faceSettings[currentFace].swStartTime = (tv.tv_sec * 100) + (tv.tv_usec / 10000);
				faceSettings[currentFace].swStartTime -= faceSettings[currentFace].swRunTime;
				faceSettings[currentFace].swRunTime = 0;
				stopwatchActive ++;
				lastTime = -1;
			}
		}
		else
		{
			if (gettimeofday(&tv, &tz) == 0)
			{
				long long tempTime = (tv.tv_sec * 100) + (tv.tv_usec / 10000);
				
				tempTime -= faceSettings[currentFace].swStartTime;		
				faceSettings[currentFace].swRunTime = tempTime;
				faceSettings[currentFace].swStartTime = -1;
				stopwatchActive --;
				lastTime = -1;
			}
		}
	}
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  S W  R E S E T  C A L L B A C K                                                                   *
 *  ===============================                                                                   *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param data .
 *  @result .
 */
static void
swResetCallback (GtkWidget * widget, guint data)
{
	if (faceSettings[currentFace].stopwatch)
	{
		int oldStartTime = faceSettings[currentFace].swStartTime;
		
		faceSettings[currentFace].swStartTime = -1;
		faceSettings[currentFace].swRunTime = 0;
		
		if (oldStartTime != -1)
		{
			swStartCallback (widget, data);
		}
		lastTime = -1;
	}
}	

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  G E T  S T O P W A T C H  T I M E                                                                 *
 *  =================================                                                                 *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param face .
 *  @result .
 */
int 
getStopwatchTime (int face)
{
	struct timeval tv;
	struct timezone tz;
		
	if (faceSettings[face].swStartTime == -1)
	{
		return faceSettings[face].swRunTime;
	}
	else
	{
		if (gettimeofday(&tv, &tz) == 0)
		{
			long long tempTime = (tv.tv_sec * 100) + (tv.tv_usec / 10000);
			tempTime -= faceSettings[face].swStartTime;
			return (int)(tempTime % (30 * 60 * 100));
		}
	}
	return 0;
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  A D D  S T R I N G  V A L U E                                                                     *
 *  =============================                                                                     *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param addBuffer .
 *  @param maxSize .
 *  @param stringFormat .
 *  @param face .
 *  @param timeNow .
 *  @result .
 */
void addStringValue (char *addBuffer, int maxSize, char *stringFormat, int face, time_t timeNow)
{
	char tempAddStr[81], tempCommand[11];
	int i = 0, j = 0;
	struct tm tm;

	localtime_r (&timeNow, &tm);
	tempCommand[0] = tempAddStr[0] = 0;

	while (stringFormat[i])
	{
		if (tempCommand[0] == '%')
		{
			tempCommand[j] = stringFormat[i];
			tempCommand[j + 1] = 0;

			switch (tempCommand[j])
			{
			case '-':
			case '_':
				j = 2;
				break;
			case '#':				/* Timezone's city wrapped */
				strcpy (tempAddStr, faceSettings[face].currentTZDisp);
				tempCommand[j = 0] = 0;
				break;
			case '*':				/* Timezone's city */
				strcpy (tempAddStr, faceSettings[face].currentTZCity);
				tempCommand[j = 0] = 0;
				break;
			case '@':				/* Timezone's area */
				strcpy (tempAddStr, faceSettings[face].currentTZArea);
				tempCommand[j = 0] = 0;
				break;
			case '&':
			{
				int swTime = getStopwatchTime (face);
				sprintf (tempAddStr, "%d:%02d:%02d.%02d",
						swTime / 360000, (swTime / 6000) % 60, (swTime / 100) % 60, swTime % 100);
				tempCommand[j = 0] = 0;				
				break;
			}
			default:
				strftime (tempAddStr, 80, tempCommand, &tm);
				tempCommand[j = 0] = 0;
				break;
			}
		}
		else
		{
			if (stringFormat[i] == '%')
			{
				tempCommand[0] = '%';
				tempCommand[j = 1] = 0;
			}
			else
			{
				tempAddStr[0] = stringFormat[i];
				tempAddStr[1] = 0;
			}
		}
		if (tempAddStr[0])
		{
			if (strlen (tempAddStr) + strlen (addBuffer) < maxSize)
			{
				strcat (addBuffer, tempAddStr);	
			}
			tempAddStr[0] = 0;
		}
		i++;
	}
	return;
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  G E T  S T R I N G  V A L U E                                                                     *
 *  =============================                                                                     *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param outString1 .
 *  @param outString2 .
 *  @param maxSize .
 *  @param stringNumber .
 *  @param face .
 *  @param timeNow .
 *  @result .
 */
char *getStringValue (char *outString1, char *outString2, int maxSize, int stringNumber, int face, time_t timeNow)
{
	char stringFormat[81], tempString[101];
	int i = 0, j = 0, m = 0;
	
	tempString[0] = 0;
	strcpy (stringFormat, displayString[stringNumber]);
	addStringValue (tempString, 100, stringFormat, face, timeNow);
	
	if (outString2 == NULL)
	{
		strncpy (outString1, tempString, maxSize);
		outString1[maxSize] = 0;
	}
	else
	{
		outString1[0] = outString2[0] = 0;
		
		while (tempString[i] && m < 2)
		{
			if (tempString[i] == '\n')
			{
				m ++;
				j = 0;
			}
			else if (m == 0 && j < maxSize)
			{
				outString1[j++] = tempString[i];
				outString1[j] = 0;
			}
			else if (m == 1 && j < maxSize)
			{
				outString2[j++] = tempString[i];
				outString2[j] = 0;
			}
			i++;
		}
	}

	return (outString1);
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  E X P O S E  C A L L B A C K                                                                      *
 *  ============================                                                                      *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @param event .
 *  @param data .
 *  @result .
 */
gboolean
exposeCallback (GtkWidget * widget, GdkEventExpose * event, gpointer data)
{
	clockExpose (widget);
	return TRUE;
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  X  S I N  C O S                                                                                   *
 *  ===============                                                                                   *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param number .
 *  @param angle .
 *  @param useCos .
 *  @result .
 */
int xSinCos (int number, int angle, int useCos)
{
	while (angle < 0) angle += 360;
	angle %= 360;
	
	if (useCos)
		return (int)rint(number * cosTable[angle]);
		
	return (int)rint(number * sinTable[angle]);
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  C H E C K  F O R  A L A R M                                                                       *
 *  ===========================                                                                       *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param face .
 *  @param hour .
 *  @param min .
 *  @result .
 */
void checkForAlarm (int face, int hour, int min)
{
	if (faceSettings[face].alarmInfo.showAlarm)
	{
		if (faceSettings[face].alarmInfo.alarmHour == hour &&
				faceSettings[face].alarmInfo.alarmMin == min)
		{
			if (!faceSettings[face].alarmInfo.alarmShown)
			{
				GtkWidget* dialog;
				
				faceSettings[face].alarmInfo.alarmShown = 1;
				dialog = gtk_message_dialog_new_with_markup (GTK_WINDOW (mainWindow),
						GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
						GTK_MESSAGE_INFO, GTK_BUTTONS_OK,
						"%d:%02d - <b>\"%s\"</b>", hour, min,
						faceSettings[face].alarmInfo.message);
				gtk_window_set_title (GTK_WINDOW (dialog), "TzClock Alarm");
#ifdef STATUS_ICON
				if (alarmCounter == 0)
				{
					gtk_status_icon_set_visible (statusIcon, 1);
					gtk_status_icon_set_blinking (statusIcon, 1);
					gtk_status_icon_set_tooltip (statusIcon, faceSettings[face].alarmInfo.message);
				}
#endif
				gdk_beep();
				alarmCounter ++;
				
				gtk_window_set_keep_above (GTK_WINDOW (dialog), true);
				gtk_dialog_run (GTK_DIALOG (dialog));
				gtk_widget_destroy (dialog);
				alarmCounter --;
#ifdef STATUS_ICON
				if (alarmCounter == 0)
				{
					gtk_status_icon_set_blinking (statusIcon, 0);				
					gtk_status_icon_set_visible (statusIcon, 0);
				}
#endif
			}
       	}
       	else if (faceSettings[face].alarmInfo.alarmShown)
       	{
       		faceSettings[face].alarmInfo.alarmShown = 0;
       	}
	}
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  L O A D  C O L O U R                                                                              *
 *  ====================                                                                              *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param fromColour .
 *  @result .
 */
static void loadColour (char *fromColour)
{
	int i, colour = -1;
	static char readColour[10], allowedChar[20];

	strcpy (readColour, "#000000");
	strcpy (allowedChar, "01234567890ABCDEF");
	
	if (strlen (fromColour) == 10)
	{
		if (fromColour[3] == '#')
		{
			for (i = 2; i < MAX__COLOURS; i++)
			{
				if (!strncmp (fromColour, nameColours[i], 3))
				{
					colour = i;
					break;
				}
			}			
			if (colour != -1)
			{
				for (i = 4; i < 10; i++)
				{
					if (strchr(allowedChar, fromColour[i]))
						readColour[i - 3] = fromColour[i];
				}
				if (i == 10)
				{
					readColour[7] = 0;
					strcpy (delaultColour[colour], readColour);
					return;
				}
			}
		}
	}
	howTo (stderr, "ERROR: Invalid colour format specified (%s)\n", fromColour);
	exit (0);
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  L O A D  D A T E  F O R M A T                                                                     *
 *  =============================                                                                     *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param newFormat .
 *  @result .
 */
void loadDateFormat (char *newFormat)
{
	int i = 0, j = 0, format = -1;
	char foundStr[61];

	if (strlen (newFormat) >= 4)
	{
		if (newFormat[3] == ':')
		{
			for (i = 0; i < TXT_COUNT; i++)
			{
				if (!strncmp (newFormat, nameFormats[i], 3))
				{
					format = i;
					break;
				}
			}			
			if (format != -1)
			{
				i = 4;
				foundStr[j] = 0;
				
				while (newFormat[i] >= ' ' && j < 60)
				{
					foundStr[j++] = newFormat[i++];
					foundStr[j] = 0;
				}
				strcpy (displayString[format], foundStr);
				return;
			}
		}
	}
	howTo (stderr, "ERROR: Invalid date format specified (%s)\n", newFormat);
	exit (0);
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  L O A D  A L A R M  I N F O                                                                       *
 *  ===========================                                                                       *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param face .
 *  @param buff .
 *  @result .
 */
void loadAlarmInfo (int face, char *buff)
{
	int alHour = 0, alMin = 0, i = 0, j = 0, m = 0;
	char msg[41];
		
	while (buff[i] && m < 3)
	{
		if (buff[i] == ':')
		{
			m++;
		}
		else if (m == 0 && buff[i] >= '0' && buff[i] <= '9')
		{
			alHour *= 10;
			alHour += (buff[i] - '0');
		}
		else if (m == 1 && buff[i] >= '0' && buff[i] <= '9')
		{
			alMin *= 10;
			alMin += (buff[i] - '0');
		}
		else if (m == 2 && j < 40)
		{
			msg[j++] = buff[i];
			msg[j] = 0;
		}
		i ++;
	}
	if (msg[0] && alHour < 24 && alMin < 60)
	{
		faceSettings[face].alarmInfo.showAlarm = 1;
		faceSettings[face].alarmInfo.alarmHour = alHour;
		faceSettings[face].alarmInfo.alarmMin = alMin;
		strcpy (faceSettings[face].alarmInfo.message, msg);
	}
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  P R O C E S S  C O M M A N D  L I N E                                                             *
 *  =====================================                                                             *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param argc .
 *  @param argv .
 *  @param posX .
 *  @param posY .
 *  @result .
 */
void processCommandLine (int argc, char *argv[], int *posX, int *posY)
{
	int i, j, face = 0, invalidOption = 0;
	char cityName[25];

	for (i = 1; i < argc; i++)
	{
		if (argv[i][0] == '-')
		{
			switch (argv[i][1])
			{
			case 'A':
				loadAlarmInfo (face, &argv[i][2]);
				break;
			case 'a':							/* Set the clock always on top */
				alwaysOnTop = !alwaysOnTop;
				break;
			case 'b':							/* Select sub-second hand */
				faceSettings[face].subSecond = !faceSettings[face].subSecond;
				break;
			case 'c':							/* Select the colours */
				loadColour (&argv[i][2]);
				break;			
			case 'd':							/* Set the date format */
				loadDateFormat (&argv[i][2]);	
				break;
			case 'f':							/* Select the face for zone changes */
				{
					int f = atoi (&argv[i][2]);
					if (f > 0 && f <= MAX_FACES)
					{
						face = f - 1;
						currentFace = face;
					}
					else
						invalidOption = 1;
				}
				break;
			case 'F':
				fastSetting = !fastSetting;
				break;
			case 'h':							/* Select sub-second hand */
				faceSettings[face].showSeconds = !faceSettings[face].showSeconds;
				break;
			case 'i':							/* Use clock icon from the theme */
				useThemeIcon = !useThemeIcon;
				break;				
			case 'l':							/* Lock the clocks position */
				lockMove = !lockMove;
				break;
			case 'n':							/* Set the number of ... */
				switch (argv[i][2])
				{
				case 'c':						/* ... columns */
					{
						int c = atoi (&argv[i][3]);
						if (c > 0 && c <= 10 && c * faceHeight <= MAX_FACES)
							faceWidth = c;
						else
							invalidOption = 1;
					}
					break;
				case 'r':						/* ... rows */
					{
						int r = atoi (&argv[i][3]);
						if (r > 0 && r <= 10 && r * faceWidth <= MAX_FACES)
							faceHeight = r;
						else
							invalidOption = 1;
					}
					break;
				default:
					invalidOption = 1;
					break;
				}
				break;
			case 'S':
				faceSettings[face].stopwatch = !faceSettings[face].stopwatch;
				break;
			case 's':							/* Select the faceSize of the clock */
				faceSize = atoi (&argv[i][2]);
				break;
			case 'T':							/* Force the clock to show a fixed time */
				forceTime = atoi (&argv[i][2]);
				break;
			case 'u':							/* Uppercase the city name */
				faceSettings[face].upperCity = !faceSettings[face].upperCity;
				break;
			case 'w':							/* Show on all workspaces */
				stuckOnAll = !stuckOnAll;
				break;
			case 'x':							/* Set the x position for the clock */
				if (argv[i][2] == 'c')
					*posX = -2;
				else if (argv[i][2] == 'r')
					*posX = -3;
				else
				{
					int x = atoi (&argv[i][2]);
					if (x >= 0)
						*posX = x;
				}
				break;
			case 'y':							/* Set the y position for the clock */
				if (argv[i][2] == 'c')
					*posY = -2;
				else if (argv[i][2] == 'b')
					*posY = -3;
				else
				{
					int y = atoi (&argv[i][2]);
					if (y >= 0)
						*posY = y;
				}
				break;
			case 'z':							/* Select which timezone to show */
				{
					for (j = 0; j < nTimeZones; j++)
					{
						splitTimeZone (timeZones[j].envName, NULL, cityName, NULL, 0);
						if (strcasecmp (&argv[i][2], cityName) == 0)
						{
							faceSettings[face].currentTZ = j;
							break;
						}
					}
					if (j == nTimeZones)
						invalidOption = 1;
				}
				break;
			case '2':
				if (argv[i][2] == '4')
				{
					faceSettings[face].show24Hour = !faceSettings[face].show24Hour;
				}
				else
				{
					invalidOption = 1;
				}
				break;
			case '?':							/* Display the help information */
				howTo (stderr, NULL);
				exit (0);
			default:
				howTo (stderr, "ERROR: Unrecognized option '%s'\n", argv[i]);
				exit (0);
			}
		}
		else
			invalidOption = 1;
			
		if (invalidOption)
		{
			howTo (stderr, "ERROR: Please check option '%s'\n", argv[i]);
			exit (0);
		}
	}
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  M A I N                                                                                           *
 *  =======                                                                                           *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param argc .
 *  @param argv .
 *  @result .
 */
int
main (int argc, char *argv[])
{
	int posX = -1, posY = -1;
	int fullHeight, fullWidth, saveFace, i;
	GtkIconTheme *iconTheme;
	GdkPixbuf *themeIcon;
	GtkWidget *eventBox;

	for (i = 0; i < MAX_FACES; i++)
	{
		faceSettings[i].showSeconds = 1;
	}
	processCommandLine (argc, argv, &posX, &posY);
	saveFace = currentFace;
	for (i = 0; i < (faceHeight * faceWidth); i++)
	{
		currentFace = i;
		setTimeZoneCallback (NULL, faceSettings[i].currentTZ);
		faceSettings[i].swStartTime = -1;
	}
	currentFace = saveFace;
	fillSinCosTables ();

	/*------------------------------------------------------------------------------------------------*
	 * Initalaise the window.                                                                         *
	 *------------------------------------------------------------------------------------------------*/
	gtk_init (&argc, &argv);
	mainWindow = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	
	/*------------------------------------------------------------------------------------------------*
	 * This is the first time we can do this because we check the screen size in this routine.        *
	 *------------------------------------------------------------------------------------------------*/
	fixFaceSize ();
	fullHeight = faceHeight * faceSize;
	fullWidth = faceWidth * faceSize;

	/*------------------------------------------------------------------------------------------------*
	 * Make a mask for the clock face.                                                                *
	 *------------------------------------------------------------------------------------------------*/
	createColours();
	makeWindowMask ();

	/*------------------------------------------------------------------------------------------------*
	 * Do all the other windows initialisation.                                                       *
	 *------------------------------------------------------------------------------------------------*/
	gtk_window_set_resizable (GTK_WINDOW (mainWindow), FALSE);
	gtk_window_set_title (GTK_WINDOW (mainWindow), "TzClock");

	menuBar = createMenus (mainWindow, GTK_TYPE_MENU);

	/*------------------------------------------------------------------------------------------------*
	 * Icon stuff.                                                                                    *
	 *------------------------------------------------------------------------------------------------*/
	if (useThemeIcon)
	{
		iconTheme = gtk_icon_theme_get_default();
		themeIcon = gtk_icon_theme_load_icon (iconTheme, "clock", 48, 0, NULL);
		if (themeIcon)
		{
			defaultIcon = gdk_pixbuf_copy (themeIcon);
	    	g_object_unref (themeIcon);
	  	}
	}
	if (!defaultIcon)
	{
		defaultIcon = gdk_pixbuf_new_from_xpm_data ((const char **) &TzClockIcon);
	}
	gtk_window_set_default_icon(defaultIcon);
#ifdef STATUS_ICON
	statusIcon = gtk_status_icon_new_from_pixbuf (defaultIcon);
	gtk_status_icon_set_visible (statusIcon, 0);
#endif

	/*------------------------------------------------------------------------------------------------*
	 * Final windows configuration.                                                                   *
	 *------------------------------------------------------------------------------------------------*/
	drawingArea = gtk_drawing_area_new ();
	gtk_widget_set_size_request (drawingArea, fullWidth, fullHeight);

	g_signal_connect (G_OBJECT (drawingArea), "expose_event", G_CALLBACK (exposeCallback), NULL);
	g_signal_connect (G_OBJECT (mainWindow), "button_press_event", G_CALLBACK (windowClickCallback), NULL);
	g_signal_connect (G_OBJECT (mainWindow), "key_press_event", G_CALLBACK (windowKeyCallback), NULL);
	g_signal_connect (G_OBJECT (mainWindow), "key_release_event", G_CALLBACK (windowKeyCallback), NULL);
	g_signal_connect (G_OBJECT (mainWindow), "destroy", G_CALLBACK (quitCallback), NULL);

	g_signal_connect (G_OBJECT (mainWindow), "focus-in-event", G_CALLBACK(focusInEvent), NULL);
	g_signal_connect (G_OBJECT (mainWindow), "focus-out-event", G_CALLBACK(focusOutEvent), NULL);
#ifdef STATUS_ICON	
	g_signal_connect (G_OBJECT (statusIcon), "activate", G_CALLBACK(statusActivate), NULL);
#endif
 	eventBox = gtk_event_box_new ();

	gtk_container_add (GTK_CONTAINER (eventBox), drawingArea);
	gtk_container_add (GTK_CONTAINER (mainWindow), eventBox);

	gtk_window_set_decorated (GTK_WINDOW (mainWindow), FALSE);

	/*------------------------------------------------------------------------------------------------*
	 * Complete stuff left over from the command line                                                 *
	 *------------------------------------------------------------------------------------------------*/
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM (
			gtk_item_factory_get_item (itemFactory, PREF_DESKTOP)),	stuckOnAll != 0);
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM (
			gtk_item_factory_get_item (itemFactory, PREF_LOCK)), lockMove != 0);
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM (
			gtk_item_factory_get_item (itemFactory, PREF_ONTOP)),	alwaysOnTop != 0);

	if (posX != -1 && posY != -1)
	{
		if (posX == -2)
			posX = (gdk_screen_width() - fullWidth) / 2;
		if (posY == -2)
			posY = (gdk_screen_height() - fullHeight) / 2;
		if (posX == -3)
			posX = gdk_screen_width() - fullWidth;
		if (posY == -3)
			posY = gdk_screen_height() - fullHeight;
		if (posX > gdk_screen_width() - 64)
			posX = gdk_screen_width() - 64;
		if (posY > gdk_screen_height() - 64)
			posY = gdk_screen_height() - 64;

		gtk_window_move (GTK_WINDOW (mainWindow), posX, posY);
	}

	/*------------------------------------------------------------------------------------------------*
	 * OK all ready lets run it!                                                                      *
	 *------------------------------------------------------------------------------------------------*/
	gtk_widget_show_all (mainWindow);
	g_timeout_add (50, clockTickCallback, NULL);

	gtk_main ();
	exit (0);
}
