/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  T z C L O C K C A I R O. C                                                                        *
 *  ==========================                                                                        *
 *                                                                                                    *
 *  TzClock developed by Chris Knight based on glock by Eric L. Smith.                                * 
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  TzClock is free software; you can redistribute it and/or modify it under the terms of the GNU     *
 *  General Public License version 2 as published by the Free Software Foundation.  Note that I       *
 *  am not granting permission to redistribute or modify TzClock under the terms of any later         *
 *  version of the General Public License.                                                            *
 *                                                                                                    *
 *  TzClock is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without      *
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 *  GNU General Public License for more details.                                                      *
 *                                                                                                    *
 *  You should have received a copy of the GNU General Public License along with this program (in     * 
 *  the file "COPYING"); if not, write to the Free Software Foundation, Inc.,                         *
 *  59 Temple Place - Suite 330, Boston, MA 02111, USA.                                               *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @file
 *  @brief .
 */
 
#include "config.h"
#include "TzClockDisp.h"
#include "TimeZone.h"

extern GtkWidget *mainWindow;
extern GtkWidget *drawingArea;
extern GdkColor clockColours[];
extern int faceSize;
extern int faceWidth;
extern int faceHeight;
extern int weHaveFocus;
extern int currentFace;
extern int timeSetting;
extern int fastSetting;
extern time_t forceTime;

extern FACE_SETTINGS faceSettings[];
extern tzInfo timeZones[];

static GdkDrawable *windowShapeBitmap;
static char windowTitle[128];

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  M A K E  W I N D O W  M A S K                                                                     *
 *  =============================                                                                     *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @result .
 */
void makeWindowMask ()
{
	int fullHeight, fullWidth, i, j;
	GdkGC *gc;
        
	fullHeight = faceHeight * faceSize;
	fullWidth = faceWidth * faceSize;
        
	if (windowShapeBitmap)
	{
		g_object_unref (windowShapeBitmap);
		windowShapeBitmap = NULL;
	}

	windowShapeBitmap = (GdkBitmap *) gdk_pixmap_new (NULL, fullWidth, fullHeight, 1);

	gc = gdk_gc_new (windowShapeBitmap);
	gdk_gc_set_foreground (gc, &clockColours[BLACK_COLOUR]);
	gdk_gc_set_background (gc, &clockColours[WHITE_COLOUR]);

	gdk_draw_rectangle (windowShapeBitmap, gc, TRUE, 0, 0, fullWidth, fullHeight);

	gdk_gc_set_foreground (gc, &clockColours[WHITE_COLOUR]);
	gdk_gc_set_background (gc, &clockColours[BLACK_COLOUR]);

	for (i = 0; i < faceWidth; i++)
	{
		for (j = 0; j < faceHeight; j++)
			gdk_draw_arc (windowShapeBitmap, gc, TRUE, (i * faceSize) - 1, (j * faceSize) - 1, 
					faceSize + 1, faceSize + 1, 0, 360 * 64);
	}
	gtk_widget_shape_combine_mask (mainWindow, windowShapeBitmap, 0, 0);
        
	g_object_unref (gc);
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  D R A W  M I N U T E                                                                              *
 *  ====================                                                                              *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param cr .
 *  @param posX .
 *  @param posY .
 *  @param size .
 *  @param len .
 *  @param angle .
 *  @param colour .
 *  @result .
 */
void drawMinute (cairo_t *cr, int posX, int posY, int size, int len, int angle, int colour)
{
	gdk_cairo_set_source_color (cr, &clockColours[colour]);
	if (len < size)
	{
		cairo_move_to (cr,
				posX + xSinCos ((faceSize * size) / 64, angle, 0),
				posY - xSinCos ((faceSize * size) / 64, angle, 1));

		cairo_line_to (cr,
				posX + xSinCos ((faceSize * (size + len)) / 64, angle, 0),
				posY - xSinCos ((faceSize * (size + len)) / 64, angle, 1));
	}
	else
	{
		cairo_move_to (cr,
				posX + xSinCos ((faceSize * size) / 64, angle + 180, 0),
				posY - xSinCos ((faceSize * size) / 64, angle + 180, 1));

		cairo_line_to (cr,
				posX + xSinCos ((faceSize * (size + len)) / 64, angle, 0),
				posY - xSinCos ((faceSize * (size + len)) / 64, angle, 1));
	}
	cairo_stroke (cr);
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  D R A W  C I R C L E                                                                              *
 *  ====================                                                                              *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param cr .
 *  @param posX .
 *  @param posY .
 *  @param size .
 *  @param colFill .
 *  @param colOut .
 *  @result .
 */
void drawCircle (cairo_t *cr, int posX, int posY, int size, int colFill, int colOut)
{
	if (colFill != -1)
	{
		gdk_cairo_set_source_color (cr, &clockColours[colFill]);
 		cairo_arc (cr, posX, posY, (faceSize * size) / 128, 0, 2 * M_PI);
		cairo_fill (cr);
		cairo_stroke (cr);
	}
	if (colOut != -1)
	{
		gdk_cairo_set_source_color (cr, &clockColours[colOut]);
 		cairo_arc (cr, posX, posY, (faceSize * size) / 128, 0, 2 * M_PI);
		cairo_stroke (cr);
	}
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  D R A W  H A N D                                                                                  *
 *  ================                                                                                  *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param cr .
 *  @param posX .
 *  @param posY .
 *  @param size .
 *  @param angle .
 *  @param colFill .
 *  @param colOut .
 *  @result .
 */
void drawHand (cairo_t *cr, int posX, int posY, int size, int angle, int colFill, int colOut)
{
	int points[8], i;
	
	points[0] = posX + xSinCos ((faceSize * (size / 3)) / 64, angle + 180, 0);
	points[1] = posY - xSinCos ((faceSize * (size / 3)) / 64, angle + 180, 1);
	points[2] = posX + xSinCos (faceSize / 64, angle + 90, 0);
	points[3] = posY - xSinCos (faceSize / 64, angle + 90, 1);
	points[4] = posX + xSinCos ((faceSize * size) / 64, angle, 0);
	points[5] = posY - xSinCos ((faceSize * size) / 64, angle, 1);
	points[6] = posX + xSinCos (faceSize / 64, angle + 270, 0);
	points[7] = posY - xSinCos (faceSize / 64, angle + 270, 1);
	
	for (i = 0; i < 2; i++)
	{
		gdk_cairo_set_source_color (cr, &clockColours[i == 0 ? colFill : colOut]);
		cairo_move_to (cr, points[0], points[1]);
		cairo_line_to (cr, points[2], points[3]);
		cairo_line_to (cr, points[4], points[5]);
		cairo_line_to (cr, points[6], points[7]);
		cairo_close_path (cr);
		if (i == 0)
		{
			cairo_fill (cr);
		}
		cairo_stroke (cr);
	}
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  D R A W  M A R K                                                                                  *
 *  ================                                                                                  *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param cr .
 *  @param posX .
 *  @param posY .
 *  @param angle .
 *  @param size .
 *  @param colFill .
 *  @param colOut .
 *  @result .
 */
void drawMark (cairo_t *cr, int posX, int posY, int angle, int size, 
		int colFill, int colOut)
{
	int points[6], i;
	
	points[0] = posX + xSinCos ((faceSize * size) / 64, angle + 2, 0);
	points[1] = posY - xSinCos ((faceSize * size) / 64, angle + 2, 1);
	points[2] = posX + xSinCos ((faceSize * (size - 3)) / 64, angle, 0);
	points[3] = posY - xSinCos ((faceSize * (size - 3)) / 64, angle, 1);
	points[4] = posX + xSinCos ((faceSize * size) / 64, angle - 2, 0);
	points[5] = posY - xSinCos ((faceSize * size) / 64, angle - 2, 1);
	
	for (i = 0; i < 2; i++)
	{
		gdk_cairo_set_source_color (cr, &clockColours[i == 0 ? colFill : colOut]);
		cairo_move_to (cr, points[0], points[1]);
		cairo_line_to (cr, points[2], points[3]);
		cairo_line_to (cr, points[4], points[5]);
		cairo_close_path (cr);
		if (i == 0)
		{
			cairo_fill (cr);
		}
		cairo_stroke (cr);
	}
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  D R A W  T E X T                                                                                  *
 *  ================                                                                                  *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param cr .
 *  @param posX .
 *  @param posY .
 *  @param string1 .
 *  @param string2 .
 *  @result .
 */
void drawText (cairo_t *cr, int posX, int posY, char *string1, char *string2)
{
	int height;
	static cairo_text_extents_t textExtents;
	
	cairo_set_font_size (cr, 4 + ((faceSize / 64) * 4));
	gdk_cairo_set_source_color (cr, &clockColours[TEXT__COLOUR]);

	cairo_text_extents (cr, "abcdefgABCDEFG", &textExtents);
	height = textExtents.height;
	
	if (string2[0])
	{
		cairo_text_extents (cr, string1, &textExtents);
		cairo_move_to (cr, posX + ((faceSize / 2) - (textExtents.width / 2.0f)), 
				posY - (height / 2.0f));
		cairo_show_text (cr, string1);
		
		cairo_text_extents (cr, string2, &textExtents);
		cairo_move_to (cr, posX + ((faceSize / 2) - (textExtents.width / 2.0f)), 
				posY + (height / 2.0f));
		cairo_show_text (cr, string2);
	}
	else
	{
		cairo_text_extents (cr, string1, &textExtents);
		cairo_move_to (cr, posX + ((faceSize / 2) - (textExtents.width / 2.0f)), 
				posY + (height / 2.0f));
		cairo_show_text (cr, string1);
	}
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  D R A W  F A C E                                                                                  *
 *  ================                                                                                  *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param cr .
 *  @param face .
 *  @param posX .
 *  @param posY .
 *  @param t .
 *  @result .
 */
gboolean 
drawFace (cairo_t *cr, int face, int posX, int posY, time_t t)
{
	int i, j, timeZone = faceSettings[face].currentTZ, tempTime;
	int hourAngle, minuteAngle, secondAngle, swMinAngle, swSecAngle = 0, swPrtAngle;
	char tempString1[51], tempString2[51];
	struct tm tm;

	if (timeZones[timeZone].value == 0)
	{
		unsetenv ("TZ");
	}
	else if (timeZones[timeZone].value < FIRST_CITY)
	{
		setenv ("TZ", "GMT", 1);
		t += 3600 * (timeZones[timeZone].value - GMT_ZERO);
	}
	else
	{
		setenv ("TZ", timeZones[timeZone].envName, 1);
	}
	localtime (&t);
	localtime_r (&t, &tm);
	
	/*------------------------------------------------------------------------------------------------*
	 *                                                                                                *
	 *------------------------------------------------------------------------------------------------*/	
	checkForAlarm (face, tm.tm_hour, tm.tm_min);
	
	/*------------------------------------------------------------------------------------------------*
	 * Handle the smooth scroll time chage                                                            *
	 *------------------------------------------------------------------------------------------------*/	
	tempTime = (tm.tm_hour * 60) + tm.tm_min;
		
	if (!fastSetting)
	{
		int gap = 0;
		
		/*--------------------------------------------------------------------------------------------*
		 * Don't scroll at midnight                                                                   *
		 *--------------------------------------------------------------------------------------------*/
		if (tempTime == 0 && faceSettings[face].shownTime == (23 * 60) + 59)
			gap = 1;
		else
			gap = tempTime - faceSettings[face].shownTime;
		
		if (abs (gap) > 1)
		{		
			/*----------------------------------------------------------------------------------------*
			 * The maximum gap is 12 hours unless it is a 24 hour clock                               *
			 *----------------------------------------------------------------------------------------*/
			if (abs(gap) > TWELVEHOURS && !faceSettings[face].show24Hour)
			{
				if (gap > 0)
				{
					faceSettings[face].shownTime += TWELVEHOURS;
					gap -= TWELVEHOURS;
				}
				else
				{
					faceSettings[face].shownTime -= TWELVEHOURS;
					gap += TWELVEHOURS;
				}
			}

			j = 1;
			for (i = 1; i <= faceSettings[face].stepping; i ++)
			{
				j += (i / 3);
				if (j > abs (gap))
					break;
			}
			i /= 3;
			if (gap < 0)
				i *= -1;

    	    if (faceSettings[face].stepping < 24)
				faceSettings[face].stepping ++;

			tempTime = faceSettings[face].shownTime + i;
			
			tm.tm_min = tempTime % 60;
			tm.tm_hour = tempTime / 60;
			
			timeSetting = 1;
		}
		else
			faceSettings[face].stepping = 0;
	}
	faceSettings[face].shownTime = tempTime;

	/*------------------------------------------------------------------------------------------------*
	 * Draw the face, it is made up of 3 overlapping circles                                          *
	 *------------------------------------------------------------------------------------------------*/	
	cairo_save (cr);
	cairo_set_line_width (cr, 1.0f + ((float)faceSize / 256.0f));
	cairo_set_line_cap (cr, CAIRO_LINE_CAP_BUTT);
	cairo_set_line_join (cr, CAIRO_LINE_JOIN_MITER);
	
	if (weHaveFocus && face == currentFace)
		gdk_cairo_set_source_color (cr, &clockColours[FACE1_COLOUR]);
	else
		gdk_cairo_set_source_color (cr, &clockColours[FACE2_COLOUR]);

	cairo_rectangle (cr, posX, posY, faceSize, faceSize);
	cairo_fill (cr);
	drawCircle (cr, posX + (faceSize / 2), posY + (faceSize / 2), 62, FACE3_COLOUR, -1);
	drawCircle (cr, posX + (faceSize / 2), posY + (faceSize / 2), 60, FACE4_COLOUR, -1);
	
	if (faceSettings[face].showSeconds && (faceSettings[face].stopwatch || faceSettings[face].subSecond))
	{
		drawCircle (cr, posX + (faceSize / 2), posY + (3 * faceSize / 4), 21, FACE5_COLOUR, -1);
	}
	if (faceSettings[face].stopwatch)
	{
		drawCircle (cr, posX + (faceSize / 4), posY + (faceSize / 2), 21, FACE5_COLOUR, -1);			
		drawCircle (cr, posX + (3 * faceSize / 4), posY + (faceSize / 2), 21, FACE5_COLOUR, -1);
	}	
	
	/*------------------------------------------------------------------------------------------------*
	 * Add the text, ether the date or the timezone, plus an AM/PM indicator                          *
	 *------------------------------------------------------------------------------------------------*/
	getStringValue (tempString1, tempString2, 50, faceSettings[face].stopwatch ?
			(timeZone ? TXT_TOPSW_Z : TXT_TOPSW_L) : (timeZone ? TXT_TOP_Z : TXT_TOP_L), face, t);
	drawText (cr, posX, posY + (faceSize / 3), tempString1, tempString2);
	 
	if (!(faceSettings[face].showSeconds && (faceSettings[face].stopwatch || faceSettings[face].subSecond)))
	{
		getStringValue (tempString1, tempString2, 50, timeZone ? TXT_BOTTOM_Z : TXT_BOTTOM_L, face, t);	
		drawText (cr, posX, posY + ((faceSize * 2) / 3), tempString1, tempString2);
	}

	if (face == currentFace)
	{
		getStringValue (tempString1, tempString2, 50, timeZone ? TXT_TITLE_Z : TXT_TITLE_L, face, t);
		if (strcmp (windowTitle, tempString1))
		{
			strcpy (windowTitle, tempString1);
			gtk_window_set_title (GTK_WINDOW (mainWindow), windowTitle);
		}
	}

	/*------------------------------------------------------------------------------------------------*
	 * Draw the hour markers                                                                          *
	 *------------------------------------------------------------------------------------------------*/
	cairo_set_line_width (cr, 1.0f + ((float)faceSize / 512.0f));
	j = faceSettings[face].show24Hour ? 120 : 60;

	for (i = 0; i < j ; i++)
	{
		hourAngle = faceSettings[face].show24Hour ? i * 3 : i * 6;

		if (faceSize > (4 * 64))
		{
			if (!faceSettings[face].show24Hour || !(i % 2))
				drawMinute (cr, posX + (faceSize / 2), posY + (faceSize / 2), 30, 1, 
						hourAngle, MMARK_COLOUR);
		}
		if (!(i % 15))
		{
			drawMark (cr, posX + faceSize / 2 , posY + faceSize / 2, hourAngle, 31, 
					BFILL_COLOUR, BOXES_COLOUR);
		}
		else if (!(i % 5))
		{
			drawMinute (cr, posX + (faceSize / 2), posY + (faceSize / 2), 28, 2, 
					hourAngle, HMARK_COLOUR);
		}
		if (faceSettings[face].showSeconds && (faceSettings[face].stopwatch || faceSettings[face].subSecond))
		{
			if (!(i % 5))
				drawMinute (cr, posX + (faceSize / 2), posY + (3 * faceSize / 4),
						(i % 15) ? 9 : 8, (i % 15) ? 1 : 2, hourAngle, SWMIN_COLOUR);
		}
		if (faceSettings[face].stopwatch)
		{
			if (!(i % 6))
				drawMinute (cr, posX + (faceSize / 4), posY + (faceSize / 2),
						(i % 12) ? 9 : 8, (i % 12) ? 1 : 2, hourAngle, SWMIN_COLOUR);
			if (!(i % 4))
				drawMinute (cr, posX + (3 * faceSize / 4), posY + (faceSize / 2),
						(i % 20) ? 9 : 8, (i % 20) ? 1 : 2, hourAngle, SWMIN_COLOUR);
		}
	}
	cairo_set_line_width (cr, 1.0f + ((float)faceSize / 256.0f));

	/*------------------------------------------------------------------------------------------------*
	 * Draw the hands                                                                                 *
	 *------------------------------------------------------------------------------------------------*/
	secondAngle = tm.tm_sec * 6;
	minuteAngle = (tm.tm_min * 6) + (tm.tm_sec / 10);
	if (faceSettings[face].show24Hour)	
		hourAngle = (tm.tm_hour * 15) + (tm.tm_min / 4);
	else
		hourAngle = (tm.tm_hour * 30) + (tm.tm_min / 2);

	if (faceSettings[face].alarmInfo.showAlarm)
	{
		int alarmAngle;
		
		if (faceSettings[face].show24Hour)	
			alarmAngle = (faceSettings[face].alarmInfo.alarmHour * 15) + (faceSettings[face].alarmInfo.alarmMin / 4);
		else
			alarmAngle = (faceSettings[face].alarmInfo.alarmHour * 30) + (faceSettings[face].alarmInfo.alarmMin / 2);
	
		drawMinute (cr, posX + (faceSize / 2), posY + (faceSize / 2), 0, 16, alarmAngle, ALARM_COLOUR);			
	}
	if (faceSettings[face].stopwatch)
	{
		int swTime = getStopwatchTime(face);
		swMinAngle = (swTime / 6000) * 12;
		swSecAngle = ((swTime / 100) % 60) * 6;
		swPrtAngle = ((swTime % 100) * 36) / 10;

		drawMinute (cr, posX + (faceSize / 4), posY + (faceSize / 2), 0, 7, swPrtAngle, SWHND_COLOUR);			
		drawMinute (cr, posX + (3 * faceSize / 4), posY + (faceSize / 2), 0, 7, swMinAngle, SWHND_COLOUR);
		drawCircle (cr, posX + (faceSize / 4), posY + (faceSize / 2), 2, SWCCF_COLOUR, SWCC__COLOUR);			
		drawCircle (cr, posX + (3 * faceSize / 4), posY + (faceSize / 2), 2, SWCCF_COLOUR, SWCC__COLOUR);
	}
	if (faceSettings[face].showSeconds && (faceSettings[face].stopwatch || faceSettings[face].subSecond))
	{
		drawMinute (cr, posX + (faceSize / 2), posY + (3 * faceSize / 4), 0, 7, secondAngle, SEC___COLOUR);
		drawCircle (cr, posX + (faceSize / 2), posY + (3 * faceSize / 4), 2, CFILL_COLOUR, CIRC__COLOUR);
	}

	drawHand (cr, posX + faceSize / 2, posY + faceSize / 2, 19, hourAngle, HFILL_COLOUR, HOUR__COLOUR);
	drawHand (cr, posX + faceSize / 2, posY + faceSize / 2, 28, minuteAngle, MFILL_COLOUR, MIN___COLOUR);

	if (faceSettings[face].showSeconds || faceSettings[face].stopwatch)
	{
		if (faceSettings[face].stopwatch)
		{
			drawMinute (cr, posX + (faceSize / 2), posY + (faceSize / 2), 8, 20, swSecAngle, SWHND_COLOUR);			
			drawCircle (cr, posX + (faceSize / 2), posY + (faceSize / 2), 4, SWCCF_COLOUR, SWCC__COLOUR);
		}
		else if (faceSettings[face].subSecond)
		{
			drawCircle (cr, posX + (faceSize / 2), posY + (faceSize / 2), 4, HFILL_COLOUR, HOUR__COLOUR);
		}
		else
		{
			drawMinute (cr, posX + (faceSize / 2), posY + (faceSize / 2), 8, 20, secondAngle, SEC___COLOUR);			
			drawCircle (cr, posX + (faceSize / 2), posY + (faceSize / 2), 4, CFILL_COLOUR, CIRC__COLOUR);
		}
	}
	else
	{
		drawCircle (cr, posX + (faceSize / 2), posY + (faceSize / 2), 4, HFILL_COLOUR, HOUR__COLOUR);
	}
	cairo_restore (cr);
	return TRUE;
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  C L O C K  E X P O S E                                                                            *
 *  ======================                                                                            *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @result .
 */
void clockExpose (GtkWidget * widget)
{
	int i, j, face = 0;
	time_t t = time (NULL);
	cairo_t *cr = gdk_cairo_create (drawingArea -> window);

	/*------------------------------------------------------------------------------------------------*
	 * Reset the color and stuff back to the default.                                                 *
	 *------------------------------------------------------------------------------------------------*/
	timeSetting = 0;

	if (forceTime != -1)
		t = forceTime;
		
	for (j = 0; j < faceHeight; j++)
	{
		for (i = 0; i < faceWidth; i++)
			drawFace (cr, face++, (i * faceSize), (j * faceSize), t);
	}

	/*------------------------------------------------------------------------------------------------*
	 * Reset the color and stuff back to the default.                                                 *
	 *------------------------------------------------------------------------------------------------*/
	cairo_destroy (cr);
	cr = NULL;	
}

