/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  T Z  C L O C K  D I S P . H                                                                       *
 *  ===========================                                                                       *
 *                                                                                                    *
 *  TzClock developed by Chris Knight based on glock by Eric L. Smith.                                * 
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  TzClock is free software; you can redistribute it and/or modify it under the terms of the GNU     *
 *  General Public License version 2 as published by the Free Software Foundation.  Note that I       *
 *  am not granting permission to redistribute or modify TzClock under the terms of any later         *
 *  version of the General Public License.                                                            *
 *                                                                                                    *
 *  TzClock is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without      *
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 *  GNU General Public License for more details.                                                      *
 *                                                                                                    *
 *  You should have received a copy of the GNU General Public License along with this program (in     * 
 *  the file "COPYING"); if not, write to the Free Software Foundation, Inc.,                         *
 *  59 Temple Place - Suite 330, Boston, MA 02111, USA.                                               *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @file
 *  @brief .
 */
 
#include <ctype.h>
#include <math.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <locale.h>
#include <sys/time.h>
#include <time.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#ifdef GTK_ABOUT_DIALOG
#include <libgnomeui/libgnomeui.h>
#endif

#define MAX_FACES 36

/*----------------------------------------------------------------------------------------------------*
 * Stuff used for picking the colours on the clock. Add to the command line.                          *
 *----------------------------------------------------------------------------------------------------*/
#define BLACK_COLOUR	0	/* fixed dont change */
#define WHITE_COLOUR	1	/* fixed dont change */
#define FACE1_COLOUR	2
#define FACE2_COLOUR	3
#define FACE3_COLOUR	4
#define FACE4_COLOUR	5
#define FACE5_COLOUR	6
#define TEXT__COLOUR	7
#define HOUR__COLOUR	8
#define MIN___COLOUR	9
#define SEC___COLOUR	10
#define ALARM_COLOUR	11
#define CIRC__COLOUR	12
#define HMARK_COLOUR	13
#define MMARK_COLOUR	14
#define BOXES_COLOUR	15
#define BFILL_COLOUR	16
#define HFILL_COLOUR	17
#define MFILL_COLOUR	18
#define CFILL_COLOUR	19
#define SWHND_COLOUR	20
#define SWCC__COLOUR	21
#define SWCCF_COLOUR	22
#define SWMIN_COLOUR	23
#define MAX__COLOURS	24		

/*----------------------------------------------------------------------------------------------------*
 * Setting the format of the date string                                                           *
 *------------------------------------------------------------------------------------------------*/
#define TXT_TOP_L			0
#define TXT_TOP_Z			1
#define TXT_TOPSW_L			2
#define TXT_TOPSW_Z			3
#define TXT_BOTTOM_L		4
#define TXT_BOTTOM_Z		5
#define TXT_BOTTOMSW_L		6
#define TXT_BOTTOMSW_Z		7
#define TXT_TITLE_L			8
#define TXT_TITLE_Z			9
#define TXT_COPY_DT_L		10
#define TXT_COPY_DT_Z		11
#define TXT_COPY_D_L		12
#define TXT_COPY_D_Z		13
#define TXT_COPY_T_L		14
#define TXT_COPY_T_Z		15
#define TXT_COUNT			16				

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
#define FLAG_LOCK			1
#define FLAG_ON_TOP			2
#define FLAG_STICK			4
#define FLAG_TIPS			8
#define FLAG_STOPWATCH		16

#define TWELVEHOURS			12 * 60

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
typedef struct _alarm
{
	int alarmHour;
	int alarmMin;
	int alarmShown;
	int showAlarm;
	char message[41];
}
alarmTime;

typedef struct _faceSettings 
{
	int stopwatch;
	int subSecond;
	int showSeconds;
	int currentTZ;
	int swRunTime;
	int show24Hour;
	int upperCity;
	int shownTime;
	int stepping;
	long long swStartTime;
	char currentTZArea[25];
	char currentTZCity[25];
	char currentTZDisp[25];
	alarmTime alarmInfo;
}
FACE_SETTINGS;

/*----------------------------------------------------------------------------------------------------*
 * Stuff for the supported timezones, in future get them from the system.  The information here   *
 * needs to match the menues below.                                                               *
 *------------------------------------------------------------------------------------------------*/
typedef struct _tzInfo
{
	char *envName;
	int value;
} 
tzInfo;

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
void makeWindowMask ();
void clockExpose (GtkWidget * widget);
char *getStringValue (char *outString1, char *outString2, int maxSize, int stringNumber, int face, time_t timeNow);
int  xSinCos (int number, int angle, int useCos);
void checkForAlarm (int face, int hour, int min);
int  getStopwatchTime (int face);


